/*
   Gradient Edge Filter Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to create a gradient edge image filter effect using
   a pixel shader.  Just like with the luminance filter demo, this one has all the
   work done in the pixel shader.  The rendering code and vertex shader code is pretty
   basic.  This is just another effect that we can create with shaders.  This demo uses
   vertex and pixel shader version 3.0.
*/


// Demo: Gradient Edge Filter.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>

// Our vertex stucture define.
#define D3DFVF_D3DVertex (D3DFVF_XYZ | D3DFVF_TEX1)


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObject();
void RenderScene();
void ShutdownDirect3D();


// Direct3D Objects
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;
LPDIRECT3DVERTEXBUFFER9 Vertex_Buffer = NULL;
LPDIRECT3DTEXTURE9 Texture = NULL;

LPDIRECT3DVERTEXSHADER9 VertexShader = NULL;
LPDIRECT3DVERTEXDECLARATION9 VertexDeclaration = NULL;
LPDIRECT3DPIXELSHADER9 PixelShader = NULL;

// Our 3 matrices...
D3DXMATRIX WorldMatrix;
D3DXMATRIX ViewMatrix;
D3DXMATRIX ProjectionMatrix;

// Vertex stucture.
struct Vertex
{
   FLOAT x, y, z;
   FLOAT tu, tv;
};

// Shader declaration.
D3DVERTEXELEMENT9 SDec[] =
{
   {0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
	{0, 12, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
   D3DDECL_END()
};


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
       {
		   case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_CLOSE:
         case WM_DESTROY:
            PostQuitMessage(0);
            break;

            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,                     // The extended style.
                         "UGPClass",               // Window class.
                         "Gradient Edge Filter Demo - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW,     // Window style.
                          100, 100,                // X, Y coords.
                          640, 480,                // Window size.
                          NULL,                    // Handle to parent window.
                          NULL,                    // Menu.
                          hInstance,               // Handle to app instance.
                          NULL);                   // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;
   
   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		return false;

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = 640;
         Present_Parameters.BackBufferHeight = 480;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                           VertexProcessing,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   // Turn off lighting and culling.
   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
   D3D_Device->SetRenderState(D3DRS_DITHERENABLE, TRUE);
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);

   // Here we set up the object this demo will render.
   if(!InitializeObject()) return false;

   // Set the image states to get a good quality image.
	D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_ADDRESSU,  D3DTADDRESS_CLAMP);
	D3D_Device->SetSamplerState(0, D3DSAMP_ADDRESSV,  D3DTADDRESS_CLAMP);

   // With that information lets check and make sure that the hardware supports
   // vertex shader 3.0 and pixel shader 3.0.
   if(D3DCaps.VertexShaderVersion < D3DVS_VERSION(3, 0)) return false;
   if(D3DCaps.PixelShaderVersion < D3DPS_VERSION(3, 0)) return false;

   // This will declare the vertex shader declartion (how the shader is set up).
   if(D3D_Device->CreateVertexDeclaration(SDec, &VertexDeclaration) != D3D_OK)
      return false;

   // This will hold the source of the vertex shader.
	ID3DXBuffer *source = NULL;

   // Load the vertex shader into Direct3D.
	if(D3DXAssembleShaderFromFile("GradientVS.vsh", NULL, NULL, NULL, &source, NULL) != D3D_OK)
      return false;

   // Now we create the vertex shader for Direct3D to use.
   if(D3D_Device->CreateVertexShader((DWORD*)source->GetBufferPointer(),&VertexShader) != D3D_OK)
      return false;

   // We can Release this now because we are done.
   if(source != NULL)
      {
         source->Release(); 
         source = NULL; 
      }

   // Load the pixel shader into Direct3D.
	if(D3DXAssembleShaderFromFile("GradientPS.psh", NULL, NULL, NULL, &source, NULL) != D3D_OK)
      return false;

   // Now we create the Pixel shader for Direct3D to use.
   if(D3D_Device->CreatePixelShader((DWORD*)source->GetBufferPointer(),&PixelShader) != D3D_OK)
      return false;

   // We can Release this now because we are done.
   if(source != NULL)
      {
         source->Release(); 
         source = NULL; 
      }

   // Set the projection matrix.
   D3DXMatrixPerspectiveFovLH(&ProjectionMatrix, 45.0f, 640/480, 0.0f, 500.0f);
   D3D_Device->SetTransform(D3DTS_PROJECTION, &ProjectionMatrix);

   return true;
}


bool InitializeObject()
{
   // Square data for its position and texture coords.
	Vertex Polygon_Data[4] =
   {
		{-0.35f, 0.5f, 0.0f, 0.0, 0.0},
      {0.35f, 0.5f, 0.0f, 1.0, 0.0},
	   {-0.35f, -0.5f, 0.0f, 0.0, 1.0},
      {0.35f, -0.5f, 0.0f, 1.0, 1.0}
	};

   // Create the vertex buffer that will hold the polygon.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(Polygon_Data),
                                            0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &Vertex_Buffer, NULL)))
      return false;
	
   // Pointer to the buffer.
   Vertex* Vertices;

   // Lock the buffer we can write to it.
   if(FAILED(Vertex_Buffer->Lock(0, sizeof(Polygon_Data), (void**)&Vertices, 0)))
      return false;

   // Here we copy the polygon's data into the vertex buffer.
   memcpy(Vertices, Polygon_Data, sizeof(Polygon_Data));

   // Unlock when your done coping data into the buffer.
   Vertex_Buffer->Unlock();

   // Here we load the texture into the texture object.
   if(D3DXCreateTextureFromFile(D3D_Device, "PS3.bmp", &Texture) != D3D_OK)
      return false;

   return true;
}


void RenderScene()
{
   // Clear the screen.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct3D we are now going to start drawing.
   D3D_Device->BeginScene();

      // Apply texture.
      D3D_Device->SetTexture(0, Texture);

	   // This will bind the vertex data in the buffer to the Direct3D device.
      D3D_Device->SetStreamSource(0, Vertex_Buffer, 0, sizeof(Vertex));

      // Create the translation matrix.
      D3DXMatrixTranslation(&WorldMatrix, 0.0f, 0.0f, 1.0f);
   
      // Load the view matrix into ViewMatrix because we will also need that...
      D3DXMatrixIdentity(&ViewMatrix);

      // Create the model view projection matrix by combining them all together.
      D3DXMATRIX modelViewProj = WorldMatrix * ViewMatrix * ProjectionMatrix;

      // Transpose the model view matrix and we are all set.
      D3DXMatrixTranspose(&modelViewProj, &modelViewProj);

      // Load the model view projection matrix in registers c[0]-c[3].
      D3D_Device->SetVertexShaderConstantF(0, (float*)modelViewProj, 4);

      // Set how the vertex shader is defined.
      D3D_Device->SetVertexDeclaration(VertexDeclaration);

      // Set the vertex shader we loaded.
      D3D_Device->SetVertexShader(VertexShader);
   
      // Set the pixel shader we loaded.
      D3D_Device->SetPixelShader(PixelShader);

      // This will draw everything in the buffer.
      D3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
    
   // We are done drawing for this scene.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects and vertex shader.
   if(Texture != NULL) 
      {
         Texture->Release();
         Texture = NULL;
      }

   if(PixelShader != NULL)
      {
         PixelShader->Release();
         PixelShader = NULL;
      }

   if(VertexShader != NULL)
      {
         VertexShader->Release();
         VertexShader = NULL;
      }

   if(VertexDeclaration != NULL)
      {
         VertexDeclaration->Release();
         VertexDeclaration = NULL;
      }

   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }

   if(Vertex_Buffer != NULL)
      {
	      Vertex_Buffer->Release();
         Vertex_Buffer = NULL;
      }
}


// Recap:

// As you can see the vertex shader for this demo does nothing but pass the data along
// to the pixel shader.  It is very basic in what it does.  The real work is done in the
// pixel shader.  With a few pixel shader instructions we have yet another effect.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com